# Copyright (c) Meta, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from __future__ import annotations

from .task import PredictTask, RelaxationTask, TrainTask, ValidateTask

__all__ = ["TrainTask", "PredictTask", "ValidateTask", "RelaxationTask"]
